#beepy-loading {
    visibility: visible;
    background: none;
    position: fixed;
    top: 50%;
    left: 50%;
    width: 0;
    height: 0;
    z-index: 1;
}

#beepy-loading::before {
    content: '';
    display: inline-block;
    width: 50px;
    height: 50px;
    border: 3px solid rgba(172, 237, 255, 0.5);
    border-radius: 50%;
    border-top-color: #dddddd;
    animation: spin 1s ease-in-out infinite;
}

#beepy-loading > span {
    display: block;
    opacity: 0.5;
}

@keyframes spin {
    to {
        transform: rotate(360deg);
    }
}
